(* +------------------------------------------------------------------------+
   |                                                                        |
   |               Calcul de Pi, formule de Bailey-Fergison                 |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

open Numerix
module Main(E:Int_type) = struct
  module I = Infixes(E)
  module L = Lll.M(E)
  open E
  open I
  open Printf

  let zero = of_int 0
  and un   = of_int 1

                     (* +------------------------------+
                        |  Calcul de sommes de sries  |
                        +------------------------------+ *)

  (* on calcule les sommes en virgule fixe avec 10 bits de rserve
     puis on arrondit le rsultat *)

  let more_bits = 10
  let round x = (x +. 512) >> 10

  (* sum(1/16^k/(8k+i)) *)
  let sum16 n i =

    (* s = somme partielle, a = 2^n/16^k, x = 8k+i *)
    let rec loop s a x =
      if a < x then round s
      else loop (s + a/x) (a >> 4) (x +. 8)

    in loop zero (un << (n +! more_bits)) (of_int i)

  (* arctan(1/a) *)
  let arctan n a =
    let num = un << (n +! more_bits) in
    let a   = of_int(a) in
    let a2  = sqr(a)    in

    (* s = somme partielle, x = a^k *)
    let rec loop s k x =
      if nbits(x) >= n then round s
      else loop (if k mod 4 = 1 then s + num/(x*.k) else s - num/(x*.k))
	        (k +! 2) (x*a2)

    in loop (num/a) 3 (a*a2)

  (* pi = 16*artcan(1/5) - 4*arctan(1/239) *)
  let pi n = (arctan (n +! 4) 5) - (arctan (n +! 2) 239)




                         (* +-----------------------+
                            |  Affiche une matrice  |
                            +-----------------------+ *)

  let print_mat msg m =

    (*
      convertit tous les coefficients en chaines et calcule la largeur
      de chaque colonne
    *)
    let n = Array.length(m) and p  = Array.length(m.(0))      in
    let w = Array.make p 0  and mm = Array.make_matrix n p "" in
    for i = 0 to n-!1 do for j = 0 to p-!1 do
      mm.(i).(j) <- string_of m.(i).(j);
      w.(j) <- Pervasives.max w.(j) (String.length mm.(i).(j))
    done done;

    (* affiche la matrice *)
    printf "%s :\n" msg;
    for i = 0 to n-!1 do
      printf "[";
      for j = 0 to p-!1 do
	let l = 1 +! w.(j) -! String.length mm.(i).(j) in
        printf "%s%s " (String.make l ' ') mm.(i).(j)
      done;
      printf "]\n"
    done


                 (* +--------------------------------------+
                    |  Affiche une liste (vecteur,nombre)  |
                    +--------------------------------------+ *)

  let print_list msg m =

    (*
      convertit tous les coefficients en chaines et calcule la largeur
      de chaque colonne
    *)
    let p = Array.length(fst(List.hd m)) in
    let w = Array.make (p+!1) 0          in
    let f(v,x) = let vv = Array.map string_of v
		 and xx = string_of x in
                 for j = 0 to p-!1 do
		   w.(j) <- Pervasives.max w.(j) (String.length vv.(j))
		 done;
                 w.(p) <- Pervasives.max w.(p) (String.length xx);
		 (vv,xx)
    in
    let mm = List.map f m in

    (* affiche la liste *)
    printf "%s :\n" msg;
    let f(v,x) = printf "[";
                 for j = 0 to p-!1 do
		   let l = 1 +! w.(j) -! String.length v.(j) in
		   printf "%s%s " (String.make l ' ') v.(j)
		 done;
		 printf "]";
		 let l = 1 +! w.(p) -! String.length x in
		 printf " :%s%s\n" (String.make l ' ') x
    in
    List.iter f mm


                         (* +----------------------+
                            |  Changement de base  |
                            +----------------------+ *)

  let chgbase v x =

    let n = Array.length x
    and p = Array.length v.(0) in
    let y = Array.make p zero  in

    for i = 0 to n-!1 do
      for j = 0 to p-!1 do y.(j) <- y.(j) + x.(i)*v.(i).(j) done
    done;
    y

                         (* +-----------------------+
                            |  fonction principale  |
                            +-----------------------+ *)

  let main arglist =
    let rec parse_args n = function
      | [] -> n
      | "-n" :: s :: suite -> parse_args (int_of_string s) suite
      | _ -> printf "syntaxe : lll [-n n]\n"; flush stdout; exit 0
    in
    let n = parse_args 20 arglist in

    (* calcule les a.i avec n bits aprs la virgule *)
    let a = Array.make 9 zero in
    for i = 0 to 7 do a.(i) <- sum16 n (i +! 1) done;
    a.(8) <- pi n;
    printf "\ncalcul avec %d bits aprs la virgule\n" n;
    for i = 0 to 8 do printf "2^%d*a%d ~ %s\n" n (i+!1) (string_of a.(i)) done;

    (* constitue la matrice de Gram de la base initiale *)
    let g = Array.make_matrix 9 9 zero in
    for i = 0 to 8 do
      for j = 0 to i-!1 do
	g.(i).(j) <- a.(i)*a.(j);
   	g.(j).(i) <- g.(i).(j)
      done;
      g.(i).(i) <- sqr(a.(i)) + un
    done;
    print_mat "\nmatrice de Gram initiale" g;

    (* rduction LLL *)
    let v,_,u = L.lll (L.schmidt g) in
    print_mat "\nmatrice de Gram de la base rduite" (L.gram u);
    print_mat "\nmatrice de passage" v;

    (* petits vecteurs *)
    let d = let x = Array.make 9 zero in x.(1) <- un; L.norme u x in
    let l = L.petits u d in
    let l = List.map (fun (x,y) -> (chgbase v x),y) l      in
    print_list "\npetits vecteurs" l;

    (* contrle avec 2n bits *)
    let n2 = 2*!n in
    for i = 0 to 7 do a.(i) <- sum16 n2 (i +! 1) done;
    a.(8) <- pi n2;
    let f(v) =
      let s = ref(zero) in
      for j = 0 to 8 do s := !s + v.(j)*a.(j) done;
      v, !s
    in

    print_list
      (sprintf "\ncontrle de la base rduite avec %d bits aprs la virgule" n2)
      (List.map f (Array.to_list v));

    print_list
      (sprintf "\ncontrle des petits vecteurs avec %d bits aprs la virgule" n2)
      (List.map (fun (v,x) -> f(v)) l);

end (* module Main *)

module S = Numcount.Start(Main)
let _ = S.start()
