(* +------------------------------------------------------------------------+
   |                                                                        |
   |                           Rduction LLL                                |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

open Numerix
module M(E:Int_type) : sig

  type vecteur  = E.t array
  type matrice  = vecteur array

  val schmidt : matrice -> matrice
      (*
	G est la matrice de Gram d'une famille (u.0,..,u.n-1), suppose libre.
	[schmidt G] retourne la matrice U triangulaire suprieure
	exprimant (u.0,..,u.n-1) dans la base de Schmidt (s.0,..,s.n-1)
	associe :

              u.i = sum(u(i,j)/u(j,j)*s.j),
	      u(i,i) = (|s.0|..|s.i|)^2 = det(Gram(u.0,..,u.i))
      *)

  val gram : matrice -> matrice
      (* transformation rciproque de schmidt *)

  val norme : matrice -> vecteur -> E.t
      (*
	U est la matrice de Schmidt de (u.0,..,u.n-1).
	x = x.0*u.0 + .. + x.n-1*u.n-1
	[norme U x] retourne |x|^2
      *)
	
  val lll : matrice -> (matrice * matrice * matrice)
      (*
	U est la matrice de Schmidt de (u.0,..,u.n-1).
	[lll U] calcule une base (v.0,..,v.n-1) t-rduite au sens de Lovasz
	du groupe engendr par u.0,..,u.n-1 et etourne le triplet de matrices
	(A,B,C) avec :

	A = matrice de passage des u.i aux v.i : v.i = sum(A(i,j)*u.j)
	B = matrice de passage des v.i aux u.i : u.i = sum(B(i,j)*v.j)
	C = matrice de Schmidt associe  (v.0,..,v.n-1)
      *)

  val voisinage : matrice -> E.t -> vecteur -> E.t -> (vecteur * E.t) list
      (*
	U est la matrice de Schmidt de (u.0,..,u.n-1).
	q est un entier >= 1
	x = x.0*u.0 + .. + x.n-1*u.n-1
	d est un entier

	[voisinage U q x d] retourne la liste des couples (y, |x - qy|^2)
	tels que |x - qy|^2 <= d, trie par valeur croissante de |x - qy|^2.

	Il est garanti que cette liste est non vide si
	         d >= q^2/4*(u00 + u11/u00 + .. + u(n-1,n-1)/u(n-2,n-2))
      *)

  val petits : matrice -> E.t -> (vecteur * E.t) list
      (*
	U est la matrice de Schmidt de (u.0,..,u.n-1).
	d est un entier

	[petits U d] retourne la liste des couples (x, |x|^2)
	tels que 0 < |x|^2 <= d, trie par valeur croissante de |x|^2.
	Un seul des vecteurs x, -x est retourn.
      *)

end = struct

  module I = Infixes(E)
  open E
  open I

  type fraction = E.t * E.t
  type vecteur  = E.t array
  type matrice  = vecteur array

  let zero  = of_int(0)
  let un    = of_int(1)

                        (* +-------------------------+
                           |  Fonctions utilitaires  |
                           +-------------------------+ *)


  (* longueur *)
  let lg(m) = Array.length(m)

  (* arrondi de a/b *)
  let round a b = let q,r = quomod a b in if r*.2 >= b then q +. 1 else q



                         (* +----------------------+
                            |  Matrice de Schmidt  |
                            +----------------------+ *)
    (*
      Transforme une matrice de Gram en dcomposition des u.i sur les s.j.
      Cette dcomposition s'obtient en chelonnant la matrice de Gram
      par l'algorithme de Gauss-Bareiss.   
    *)

    let schmidt g =

      let n = lg(g)                  in
      let u = Array.map Array.copy g in
      let d = ref(un)                in

      for     i = 0    to n-!1 do
	for   j = i+!1 to n-!1 do
	  for k = i+!1 to n-!1 do
	    u.(k).(j) <- (u.(i).(i)*u.(k).(j) - u.(k).(i)*u.(i).(j)) / !d
	  done;
	  u.(i).(j) <- zero
	done;
	d := u.(i).(i);
	if !d <= zero then failwith "schmidt: matrice non dfinie positive"
      done;

      u (* rsultat *)

                       (* +--------------------------+
                          |  Transformation inverse  |
                          +--------------------------+ *)
    (*
      Recalcule la matrice de Gram  partir des dcompositions des u sur les s.

      Formule :
      (u.i | u.j) = sum( u(i,k)*u(j,k)/u(k,k)u(k-1,k-1), k=0..min(i,j) )

      La somme arrte  k=p est gale  (P(u.i) | P(u.j)) = (P(u.i) | u.j)
      o P est la projection orthogonale sur vect(u.0,..,u.p) donc
      c'est un rationnel de dnominateur u(p,p) car u(p,p)*P est 
      coefficients entiers. On peut donc conduire le calcul sans fractions
      comme dans Bareiss.
    *)
    let gram u =

      let n = lg(u)                  in
      let g = Array.map Array.copy u in

      for   i = n-!1 downto 1 do
	for j = i    downto 1 do
	  let p = ref(zero) in
	  for k = 0 to j-!1 do
	    p := !p*g.(k).(k) + g.(i).(k)*g.(j).(k);
	    if k > 0 then p := !p/g.(k-!1).(k-!1)
	  done;
	  g.(i).(j) <- (!p + g.(i).(j))/g.(j-!1).(j-!1)
	done
      done;

      for   i = 0    to n-!2 do
	for j = i+!1 to n-!1 do g.(i).(j) <- g.(j).(i) done
      done;

      g (* rsultat *)

                         (* +----------------------+
                            |  Norme d'un vecteur  |
                            +----------------------+ *)

    (* idem gram *)
    let norme u x =

      let n = lg(u) and s = ref(zero) in

      for i = 0 to n-!1 do
	let y = ref(zero) in
	for j = i to n-!1 do y := !y + u.(j).(i)*x.(j) done;
	s := !s*u.(i).(i) + sqr(!y);
	if i > 0 then s := !s/u.(i-!1).(i-!1)

      done;
      !s/u.(n-!1).(n-!1)


                           (* +------------------+
                              |  Algorithme LLL  |
                              +------------------+ *)

    let lll u =

      (* copie u pour ne pas modifier la matrice argument *)
      let n = lg(u)                  in
      let u = Array.map Array.copy u in

      (* initialise les matrices de passage *)
      let v = Array.make_matrix n n zero
      and w = Array.make_matrix n n zero in
      for i = 0 to n-!1 do v.(i).(i) <- un; w.(i).(i) <- un done;

      (* redressement de u.j par rapport  u.i *)
      let red i j =
	let q = round u.(j).(i) u.(i).(i) in
	if q <> zero then for k = 0 to n-!1 do
          u.(j).(k) <- u.(j).(k) - q*u.(i).(k);
          v.(j).(k) <- v.(j).(k) - q*v.(i).(k);
          w.(k).(i) <- w.(k).(i) + q*w.(k).(j)
	done
      in

      (* change de u.i et u.(i-1) *)
      let swap i =
	let j = i-!1  in

	(* transformation des colonnes i et j de u *)
	let a = u.(i).(j)
	and b = if i=!1 then un else u.(i-!2).(i-!2)
	and c = u.(i).(i)
	and d = u.(j).(j) in
	for k = j to n-!1 do
	  let x = u.(k).(j) and y = u.(k).(i) in
	  u.(k).(j) <- (a*x + b*y)/d;
	  u.(k).(i) <- (c*x - a*y)/d
	done;

	(* change les lignes i et j dans u,v et des colonnes i,j dans w *)
	let x = u.(i) in u.(i) <- u.(j); u.(j) <- x;
	let x = v.(i) in v.(i) <- v.(j); v.(j) <- x;
	for k = 0 to n-!1 do
	  let x = w.(k).(i) in w.(k).(i) <- w.(k).(j); w.(k).(j) <- x
	done
      in

      (* boucle de rduction :
	 la sous-famille (u0,..,u.k-1) est suppose LLL-rduite,
	 rduit les vecteurs suivants.
      *)
      let rec loop k = if k < n then begin

	(* redresse u.(k) par rapport  u.(k-1) *)
	red (k-!1) k;

	(* condition de Lovasz *)
	let x = if k=!1 then u.(k).(k) else u.(k).(k)*u.(k-!2).(k-!2) in
	if (x + sqr(u.(k).(k-!1)))*.4 >= sqr(u.(k-!1).(k-!1))*.3

	(* OK, continue le redressement *)
	then begin
	  for i = 0 to k-!2 do red i k done;
	  loop (k+!1)
	end

	(* u.k est trop court, change avec u.k-1 *)
	else begin
	  swap k;
	  loop (if k >! 1 then k-!1 else 1)
	end
      end
      in

      (* lance la rduction *)
      loop 1;
      (v,w,u)


                     (* +---------------------+
                        |  Explore une boule  |
                        +---------------------+ *)

    (*
      u = matrice de Schmidt
      q = facteur commun
      x = centre de la boule dcompos dans la base de Schmidt
      d = (rayon^2)*u(k,k)
      k = on cherche les CL de q*u.1, .., q*u.k
      pos = impose que la dre coordonne non nulle soit > 0
     *)
    let rec boule u q x d k pos =
      if d < zero then []
      else if k < 0 then if pos then [] else [Array.make (lg x) zero]
      else begin

	(* essaye avec yk, yk + qr, yk + 2qr, ... *)
	let rec loop yk r =
	  let z = Array.copy x in
	  for i = 0 to k do z.(i) <- z.(i) - q*yk*u.(k).(i) done;
	  let d = if k > 0 then d*u.(k-!1).(k-!1) else d in
	  let d = (d - sqr(z.(k)))/u.(k).(k) in
	  if d < zero then []
	  else begin
	    let l  = boule u q z d (k-!1) (pos & (yk = zero)) in
	    List.iter (fun y -> y.(k) <- yk) l;
	    l @ (loop (yk +. r) r)
	  end
	in

	(* lance la recherche *)
	if pos then loop zero 1
	else let yk = x.(k)/(q*u.(k).(k)) in (loop (yk+un) 1) @ (loop yk (-1))

      end

                   (* +------------------------+
                      |  Voisinage d'un point  |
                      +------------------------+ *)

    let voisinage u q x d =

      (* dcompose x sur la base de Schmidt (s.0, .., s.n-1) *)
      let n = lg(x)             in
      let y = Array.make n zero in
      for i = 0 to n-!1 do for j = 0 to i do
	y.(j) <- y.(j) + u.(i).(j)*x.(i)
      done done;

      (* lance la recherche *)
      let l = boule u q y (d*u.(n-!1).(n-!1)) (n-!1) false in

      (* calcule les distances et trie la liste *)
      let z = Array.make n zero in
      let f y =
	for i = 0 to n-!1 do z.(i) <- x.(i) - q*y.(i) done;
	y, (norme u y)
      in
      Sort.list (fun (_,a) (_,b) -> a < b) (List.map f l)

                           (* +-------------------+
                              |  Petits vecteurs  |
                              +-------------------+ *)

    let petits u d =

      let n = lg(u)             in
      let y = Array.make n zero in
      let l = boule u un y (d*u.(n-!1).(n-!1)) (n-!1) true in

      (* calcule les normes et trie la liste *)
      Sort.list (fun (_,a) (_,b) -> a < b) (List.map (fun v -> (v,norme u v)) l)
     

end (* module M *)
